// -*- mode: c++ basic-offset: 2; indent-tabs-mode: nil; -*-
#include "config.h"

#include <algorithm>
#include <iostream>

#include "general/unordered_map.hh"

#include "dirinfo.hh"
#include "FilenameHandler.hh"
#include "Query.hh"
#include "SDGTxFilenameHandler.hh"

using General::ReadWriteLock;

namespace
{
  using namespace diskCache;

  static const General::ReadWriteLock::Sync::lock_mode mode_readwrite
  = General::ReadWriteLock::Sync::WRITE;

  static const General::ReadWriteLock::Sync::lock_mode mode_readonly
  = General::ReadWriteLock::Sync::READ;

} /* namespace - anonymous */

namespace diskCache
{
  //=====================================================================
  // FilenameHandler::Hash
  //=====================================================================
  FilenameHandler::Hash::
  ~Hash( )
  {
  }

  //=====================================================================
  // FilenameHandler::IsFileData
  //=====================================================================
  FilenameHandler::IsFileData::PrivateData::
  PrivateData( )
    : s_added_file_count( 0 ),
      s_removed_file_count( 0 )
  {
  }

  FilenameHandler::IsFileData::PrivateData::
  ~PrivateData( )
  {
  }

  //=====================================================================
  // FilenameHandler::IsFileData
  //=====================================================================
  FilenameHandler::IsFileData::
  IsFileData( const DirectoryInfo& PreviousView,
              General::SharedPtr< DirectoryInfo > CurrentView )
    : s_previous_view( PreviousView ),
      s_current_view( CurrentView )
  {
  }

  //---------------------------------------------------------------------
  /// This method simply loops over the list of elements calling the
  /// virtual function to allow any search algorithm specific
  /// actions to be taken after all directory elements have been read.
  //---------------------------------------------------------------------
  void FilenameHandler::IsFileData::
  OnDirectoryClose( )
  {
    for ( private_data_container_type::iterator
            cur = s_private_data.begin( ),
            last = s_private_data.end( );
          cur != last;
          ++cur )
    {
      cur->second->OnDirectoryClose( s_current_view );
    }
  }

  INT_4U FilenameHandler::IsFileData::
  FilesAdded( ) const
  {
    INT_4U      retval( 0 );

    for ( private_data_container_type::const_iterator
            cur = s_private_data.begin( ),
            last = s_private_data.end( );
          cur != last;
          ++cur )
    {
      retval += cur->second->s_added_file_count;
    }

    return retval;
  }

  INT_4U FilenameHandler::IsFileData::
  FilesRemoved( ) const
  {
    INT_4U      retval( 0 );

    for ( private_data_container_type::const_iterator
            cur = s_private_data.begin( ),
            last = s_private_data.end( );
          cur != last;
          ++cur )
    {
      retval += cur->second->s_removed_file_count;
    }

    return retval;
  }

  //=====================================================================
  //=====================================================================
  // FilenameHandler
  //=====================================================================
  FilenameHandler::filename_handler_container   FilenameHandler::m_handlers;
  General::ReadWriteLock::Sync                  FilenameHandler::m_handlers_baton;

  FilenameHandler::
  ~FilenameHandler( )
  {
  }

  //=====================================================================
  // FilenameHandler::HashContainer
  //=====================================================================


  //---------------------------------------------------------------------
  /// Constructor
  //---------------------------------------------------------------------
  FilenameHandler::HashContainer::
  HashContainer( )
  {
  }

  //---------------------------------------------------------------------
  /// Register the information from all of the appropriate places.
  //---------------------------------------------------------------------
  void FilenameHandler::HashContainer::
  Register( ) const
  {
    std::list< hash_value_type >    hashes;

    {
      General::ReadWriteLock    l( m_handlers_baton,
                                   mode_readwrite,
                                   __FILE__, __LINE__  );
      for ( container_type::const_iterator
              cur = begin( ),
              last = end( );
            cur != last;
            ++cur )
      {
        hashes.push_back( cur->second );
      }
    }

#if ENABLE_MOUNT_POINT_CACHE_old
    for ( std::list< hash_value_type >::const_iterator
            cur = hashes.begin( ),
            last = hashes.end( );
          cur != last;
          ++cur )
    {
     (*cur)->Register( );
    }
#endif /* ENABLE_MOUNT_POINT_CACHE_old */
    
  }

  //---------------------------------------------------------------------
  /// Unregister the information from all of the appropriate places.
  //---------------------------------------------------------------------
  void FilenameHandler::HashContainer::
  Unregister( LogMessage& Errors ) const
  {
    std::list< hash_value_type >    hashes;

    {
      General::ReadWriteLock    l( m_handlers_baton,
                                   mode_readwrite,
                                   __FILE__, __LINE__  );
      for ( container_type::const_iterator
              cur = begin( ),
              last = end( );
            cur != last;
            ++cur )
      {
        hashes.push_back( cur->second );
      }
    }

#if ENABLE_MOUNT_POINT_CACHE_old
    for ( std::list< hash_value_type >::const_iterator
            cur = hashes.begin( ),
            last = hashes.end( );
          cur != last;
          ++cur )
    {
      (*cur)->Unregister( Errors );
    }
#endif /* ENABLE_MOUNT_POINT_CACHE_old */
    
  }

  void FilenameHandler::HashContainer::
  Find( const Query& SearchParams,
        QueryResults& Results ) const
  {
    std::list< hash_info_type >        hashes;

    {
      //-----------------------------------------------------------------
      // Lock to protect valuable information
      //-----------------------------------------------------------------
      General::ReadWriteLock    l( m_handlers_baton,
                                   mode_readonly,
                                   RWLOCK_TIMEOUT,
                                   __FILE__, __LINE__ );

      if ( SearchParams.Criteria( ) & Query::CRITERIA_BIT_FILENAME_HANDLER )
      {
        //---------------------------------------------------------------
        // Limit the search to a specific file name handler
        //---------------------------------------------------------------
        const_iterator pos = find( SearchParams.FNHandler( ) );

        if ( pos != end( ) )
        {
          hashes.push_back( *pos );
        }
      }
      else
      {
        //---------------------------------------------------------------
        // Search over all file name handlers.
        //---------------------------------------------------------------
        for ( const_iterator
                cur = begin( ),
                last = end( );
              cur != last;
              ++cur )
        {
          hashes.push_back( *cur );
        }
      }
    }

    Query       sp( SearchParams );

    for ( std::list< hash_info_type >::const_iterator
            cur = hashes.begin( ),
            last = hashes.end( );
          cur != last;
          ++cur )
    {
      //-----------------------------------------------------------------
      // Reset the query to only look within this file handler
      //-----------------------------------------------------------------
      sp.FNHandler( cur->first );
      cur->second->Find( sp, Results );
    }

  }
    
  //---------------------------------------------------------------------
  /// Register the information from all of the appropriate places.
  //---------------------------------------------------------------------
  size_t FilenameHandler::HashContainer::
  FileCount( ) const
  {
    size_type                           retval( 0 );
    std::list< hash_value_type >        hashes;

    {
      General::ReadWriteLock    l( m_handlers_baton,
                                   mode_readwrite,
                                   __FILE__, __LINE__  );
      for ( container_type::const_iterator
              cur = begin( ),
              last = end( );
            cur != last;
            ++cur )
      {
        hashes.push_back( cur->second );
      }
    }

    for ( std::list< hash_value_type >::const_iterator
            cur = hashes.begin( ),
            last = hashes.end( );
          cur != last;
          ++cur )
    {
      retval += (*cur)->FileCount( );
    }
    
    return retval;
  }

  void FilenameHandler::HashContainer::
  clear( )
  {
    if ( m_handlers_baton.IsLocked( mode_readwrite ) == false )
    {
      General::ReadWriteLock    l( m_handlers_baton,
                                   mode_readwrite,
                                   __FILE__, __LINE__  );

      clear( );
      return;
    }
    container_type::clear( );
  }

  void FilenameHandler::HashContainer::
  read( CacheFile& Stream, General::SharedPtr< DirectoryInfo > DirInfo )
  {
    if ( m_handlers_baton.IsLocked( mode_readwrite ) == false )
    {
      General::ReadWriteLock    l( m_handlers_baton,
                                   mode_readwrite,
                                   __FILE__, __LINE__  );

      read( Stream, DirInfo );
      return;
    }
    switch ( Stream.Version( ) )
    {
    case 0x0101:
    case 0x0100:
    case 0x0000:
      {
        //-----------------------------------------------------------------
        // Locate SDGTx entry
        //-----------------------------------------------------------------
        container_type::value_type::first_type
          key( SDGTxFilenameHandler::Key( ) );

        container_type::iterator
          cur = begin( ),
          last = end( );
        while( cur != last )
        {
          if ( key == cur->first )
          {
            break;
          }
          ++cur;
        }
        if ( cur == last )
        {
          hash_info_type        value;
          value.first = key;
          value.second = NULL;

          cur = insert( cur, value );
        }
        if ( cur->second == NULL )
        {
          cur->second = key->CreateHash( DirInfo );
        }
        cur->second->Read( Stream );
      }
      break;
    default:
      {
        std::ostringstream        msg;

        msg
          << "diskCache::FilenameHandler::HashContainer::read Unsupported version: "
          << Stream.Version( )
          ;
        throw std::runtime_error(  msg.str( ) );
      }
      break;
    } // switch - version
  }

  void FilenameHandler::HashContainer::
  write( CacheFile& Stream ) const
  {
    if ( m_handlers_baton.IsLocked( mode_readonly ) == false )
    {
      General::ReadWriteLock    l( m_handlers_baton,
                                   mode_readonly,
                                   RWLOCK_TIMEOUT,
                                   __FILE__, __LINE__  );

      write( Stream );
      return;
    }
    switch ( Stream.Version( ) )
    {
    case 0x0101:        // Support for multiple file extensions
    case 0x0100:        // Support for byte swapping
    case 0x0000:        // Origional format
      {
        //---------------------------------------------------------------
        // Locate SDGTx entry
        //---------------------------------------------------------------
        const FilenameHandler* const
          key( SDGTxFilenameHandler::Key( ) );

        for ( container_type::const_iterator
                cur = begin( ),
                last = end( );
              cur != last;
              ++cur )
        {
          if ( key == cur->first )
          {
            cur->second->Write( Stream );
            return;
          }
        }
        Stream << CacheFile::size_type( 0 );
      }
      break;
    default:
      {
        std::ostringstream        msg;

        msg
          << "diskCache::FilenameHandler::HashContainer::write Unsupported version: "
          << Stream.Version( )
          ;
        throw std::runtime_error(  msg.str( ) );
      }
      break;
    } // switch - version
  }

  //=====================================================================
  // FilenameHandler
  //=====================================================================
  bool FilenameHandler::
  IsFileCheck( const std::string& Filename,
          IsFileData& Data )
  {
    if ( m_handlers_baton.IsLocked( mode_readonly ) == false )
    {
      General::ReadWriteLock    l( m_handlers_baton,
                                   mode_readonly,
                                   RWLOCK_TIMEOUT,
                                   __FILE__, __LINE__  );

      return IsFileCheck( Filename, Data );
    }
    bool retval = false;

    for ( filename_handler_container::const_iterator
	    cur = m_handlers.begin( ),
	    last = m_handlers.end( );
	  cur != last;
	  ++cur )
    {
      retval = ( retval || (*cur)->IsFile( Filename, Data ) );
    }
    return retval;
  }

  bool FilenameHandler::
  RegisterHandler( FilenameHandler& Handler )
  {
    General::ReadWriteLock      l( m_handlers_baton,
                                   General::ReadWriteLock::WRITE,
                                   __FILE__, __LINE__ );

    m_handlers.push_back( &Handler );
    return true;
  }

  bool FilenameHandler::
  UnregisterHandler( FilenameHandler& Handler )
  {
    General::ReadWriteLock      l( m_handlers_baton,
                                   General::ReadWriteLock::WRITE,
                                   __FILE__, __LINE__ );

    m_handlers.remove( &Handler );
    return true;
  }
} /* namespace diskCache */

namespace
{
  bool init( );

  static const bool initialized = init( );

  bool
  init( )
  {   
    using namespace diskCache;

    SDGTxFilenameHandler::Register( );

    return ( initialized || true );
  }

} /* namespace - anonymous */
