#ifdef HAVE_CONFIG_H
#include "genericAPI/config.h"
#endif /* HAVE_CONFIG_H */

#include <signal.h>

#include <sstream>

#include "general/ErrorLog.hh"
#include "general/mutexlock.hh"
#include "general/TaskThread.hh"
#include "general/Timeout.hh"

#include "genericAPI/Logging.hh"
#include "genericAPI/MountPointStatus.hh"
#include "genericAPI/Stat.hh"
#include "genericAPI/StatFork.hh"
#include "genericAPI/StatDirect.hh"
#include "genericAPI/StatPool.hh"


using General::StdErrLog;
using General::TaskThread;
using General::Timeout;
using GenericAPI::MountPointStatus;
using GenericAPI::StatPool;
using GenericAPI::StatDirect;

// Timeout value for lstat
static int lstat_timeout_value = 5;

namespace
{
  void
  wakeup( int SignalNumber )
  {
    signal( SignalNumber, wakeup );
  }
}

namespace
{
  class timed_stat
    : public General::Task
  {
  public:
    enum mode_type {
      MODE_STAT,
      MODE_LSTAT
    };

    timed_stat( const std::string& Path, struct stat& StatBuf, mode_type Mode )
      : Task( GenericAPI::lstat_interupt( ) ),
	m_mode( Mode ),
	m_path( Path ),
	m_statbuf( StatBuf ),
	m_err( -1 ),
	m_errno( EINTR )
	
    {
    }

    inline int Error( ) const
    {
      return m_err;
    }
    
    inline int ErrNo( ) const
    {
      return m_errno;
    }

    virtual void operator()( )
    {
      static const CHAR* method_name( "General::timed_stat::operator()" );

      if ( StdErrLog.IsOpen( ) )
      {
	std::ostringstream	msg;
      
	msg << "lstat_timed: thread: " << (void*)pthread_self( )
	    << " lstat(" << m_path << ")";
	StdErrLog( General::ErrorLog::DEBUG,
		   __FILE__, __LINE__,
		   msg.str( ) );
	struct timespec wakeup;

	wakeup.tv_sec = 1;
	wakeup.tv_nsec = 0;
	m_err = ( nanosleep( &wakeup, NULL ) != 0 );
      }

      m_stat = StatPool::Request( );
      
      try
      {
	switch( m_mode )
	{
	case MODE_STAT:
	  m_err = m_stat->Stat( m_path, m_statbuf );
	  break;
	case MODE_LSTAT:
	  m_err = m_stat->LStat( m_path, m_statbuf );
	  break;
	default:
	  m_errno = -1;
	  break;
	}
      }
      catch( ... )
      {
	m_errno = -1;
	try
	{
	  throw;
	}
	catch( const General::TimeoutException& Exception )
	{
	  std::ostringstream   msg;

	  msg << "TimeoutException for: " << m_path
	      << " (" << Exception.what( )
	      << ")"
	    ;
   
	  GenericAPI::queueLogEntry( msg.str( ),
				     GenericAPI::LogEntryGroup_type::MT_DEBUG,
				     0,
				     method_name,
				     "CXX_INTERFACE" );
	  MountPointStatus::Offline( m_path );
	  if ( StdErrLog.IsOpen( ) )
	  {
	    StdErrLog( General::ErrorLog::DEBUG,
		       __FILE__, __LINE__,
		       Exception.what( ) );
	  }
	}
	catch( const std::exception& Exception )
	{
	  std::ostringstream   msg;

	  msg << "std::exception for: " << m_path
	      << " (" << Exception.what( )
	      << ")"
	    ;
   
	  GenericAPI::queueLogEntry( msg.str( ),
				     GenericAPI::LogEntryGroup_type::MT_DEBUG,
				     0,
				     method_name,
				     "CXX_INTERFACE" );
	  if ( StdErrLog.IsOpen( ) )
	  {
	    StdErrLog( General::ErrorLog::DEBUG,
		       __FILE__, __LINE__,
		       Exception.what( ) );
	  }
	}
	catch( ... )
	{
	  std::ostringstream   msg;

	  msg << "unknown exception for: " << m_path
	    ;
   
	  GenericAPI::queueLogEntry( msg.str( ),
				     GenericAPI::LogEntryGroup_type::MT_DEBUG,
				     0,
				     method_name,
				     "CXX_INTERFACE" );
	  // :TODO: Unknown exception
	}
      }
      m_errno = errno;
      StatPool::Release( m_stat );
      m_stat = (StatPool::info_type)NULL;
      if ( m_err != 0 )
      {
	std::ostringstream	msg;
      
	msg << "lstat_timed: thread: " << (void*)pthread_self( )
	    << " m_err: " << m_err
	    << " m_errno: " << m_errno;

	GenericAPI::queueLogEntry( msg.str( ),
				   GenericAPI::LogEntryGroup_type::MT_DEBUG,
				   0,
				   method_name,
				   "CXX_INTERFACE" );
      }
    }

  private:
    const mode_type	m_mode;
    const std::string	m_path;
    struct stat&	m_statbuf;
    int			m_err;
    int			m_errno;
    StatPool::info_type	m_stat;

    static void
    destroyer( timed_stat* TStat )
    {
      volatile StatPool::info_type rstat = TStat->m_stat;

      TStat->m_stat = (StatPool::info_type)NULL;
      //-----------------------------------------------------------------
      // This is called when things are taking too long
      //-------------------------------------------------------------------
      if ( StdErrLog.IsOpen( ) )
      {
	std::ostringstream	msg;
	
	msg << "destroyer: Stat buffer: " << (void*)rstat;
	StdErrLog( General::ErrorLog::DEBUG,
		   __FILE__, __LINE__,
		   msg.str( ) );
	struct timespec wakeup;
      
	wakeup.tv_sec = 1;
	wakeup.tv_nsec = 0;
	nanosleep( &wakeup, NULL );
      }
      StatPool::Destroy( rstat );
      TStat->m_err = -1;
      TStat->m_errno = EINTR;
    }
  };

} // namespace - anonymous

namespace GenericAPI
{
  int
  init_lstat_interupt( )
  {
    // static const int interupt = SIGTERM;
    static const int interupt = TaskThread::CANCEL_SIGNAL_ABANDON;
    static bool initialized = false;

    if ( initialized == false )
    {
      static MutexLock::lock_type baton = PTHREAD_MUTEX_INITIALIZER;

      MutexLock l( baton );

      if ( initialized == false )
      {
	const StatPool::info_type lstat_base = StatPool::StatType( );
    
	if ( ( interupt > 0 )
	     && ( dynamic_cast< const StatDirect* >( lstat_base ) ) )
	{
	  wakeup( interupt );	// Initialize the signal handler
	}
	initialized = true;
      }
    }

    return interupt;
  }

  int
  SetStatTimeout( int Value )
  {
    static const CHAR* const
      method_name( "GenericAPI::SetStatTimeout" );

    int	retval = lstat_timeout_value;

    std::ostringstream	msg;

    msg << "Modifying lstat timeout from: " << retval
	<< " to: " << Value
      ;
    queueLogEntry( msg.str( ),
		   GenericAPI::LogHTML::MT_DEBUG,
		   0,
		   method_name,
		   "VARIABLE" );
    
    lstat_timeout_value = Value;

    return retval;
  }

  int
  Stat( const std::string& Filename, struct stat& StatBuffer )
  {
    int retval = -1;

    if ( MountPointStatus::Status( Filename )
	 == MountPointStatus::STATE_OFFLINE )
    {
      errno = EINTR;
      return retval;
    }

    //-------------------------------------------------------------------
    // PR#3015 -
    // Be careful with the lstat call since it can lock up under
    //   nfs usage and not return until the server is once again
    //   available.
    // This condition is avoided by setting a maximum time for this
    //   call to return before causing the lstat call to be interupted.

    timed_stat*	tstat
      = new timed_stat( Filename, StatBuffer, timed_stat::MODE_STAT );

    if ( Timeout( tstat, lstat_timeout( ), false ) == 0 )
    {

      retval = tstat->Error( );
      errno = tstat->ErrNo( );
      delete tstat;
    }
    else
    {
      MountPointStatus::Offline( Filename );
      errno = EINTR;
    }

    return retval;
  }

  int
  LStat( const std::string& Filename, struct stat& StatBuffer )
  {
    int retval = -1;

    if ( MountPointStatus::Status( Filename )
	 == MountPointStatus::STATE_OFFLINE )
    {
      errno = EINTR;
      return retval;
    }

    //-------------------------------------------------------------------
    // PR#3015 -
    // Be careful with the lstat call since it can lock up under
    //   nfs usage and not return until the server is once again
    //   available.
    // This condition is avoided by setting a maximum time for this
    //   call to return before causing the lstat call to be interupted.
    //-------------------------------------------------------------------
    timed_stat*	tstat
      = new timed_stat( Filename, StatBuffer, timed_stat::MODE_LSTAT );

    if ( Timeout( tstat, lstat_timeout( ), false ) == 0 )
    {
      retval = tstat->Error( );
      errno = tstat->ErrNo( );
      delete tstat;
    }
    else
    {
      MountPointStatus::Offline( Filename );
      errno = EINTR;
    }

    return retval;
  }

  int
  lstat_timeout( )
  {
    return lstat_timeout_value;
  }
}

